import numpy
import ipywidgets as widgets
from IPython.display import display, clear_output
import plotly.graph_objs as go
import pandas as pd


class ManualGame:
    def __init__(self, team_names, max_rounds, cooldown, width=1000):
        self.team_names = team_names
        self.max_rounds = max_rounds
        self.cooldown = cooldown
        self.gen = numpy.random.default_rng(seed=None)
        self.width = width
        self.initialize_ui()

    def initialize_ui(self):
        self.current_round = 0
        self.team_scores = {team: [0] for team in self.team_names}
        self.team_blocked_until = {team: -1 for team in self.team_names}
        self.last_number = None
        self.last_cooldown = self.cooldown
        self.number_display = widgets.HTML(layout=dict(width='80%'))
        self.team_buttons = {
            team: widgets.ToggleButton(description=team, value=False, tooltip=team, layout=dict(width='50%'))
            for team in self.team_names
        }
        self.submit_button = widgets.Button(description="Submit Selections", button_style='success', layout=dict(width='50%'))
        self.output = widgets.Output()
        self.controls = widgets.VBox([
            self.number_display,
            *self.team_buttons.values(),
            self.submit_button,
            self.output,
        ],
            layout=dict(width='40%', align_items='center')
        )
        self.plot_output = widgets.Output(layout=dict(width='100%'))

        def update_plot():
            fig = go.Figure()
            for team in self.team_names:
                fig.add_trace(go.Scatter(
                    x=list(range(self.current_round + 1)),
                    y=self.team_scores[team],
                    mode='lines+markers',
                    name=team
                ))
            fig.update_layout(xaxis_title='Round', yaxis_title='Score', xaxis_range=[0, self.max_rounds], width=self.width)
            with self.plot_output:
                clear_output(wait=True)
                display(fig)

        def start_new_round():
            self.current_round += 1
            self.last_number = self.gen.pareto(3)

            self.number_display.value = f"""
                <div style='text-align:center; padding: 10px;'>
                    <h1 style='margin-bottom: 0;'>Round {self.current_round} / {self.max_rounds} <br> Cooldown: {self.last_cooldown}</h1>
                    <h2 style='color:darkred; font-size: 48px; margin-top: 10px;'>Reward: {self.last_number:.2f}</h2>
                </div>
            """

            for team, btn in self.team_buttons.items():
                blocked = self.team_blocked_until[team] - self.current_round
                is_blocked = blocked >= 0
                btn.disabled = is_blocked
                btn.value = False
                btn.description = f'❌ {team} ({blocked + 1})' if is_blocked else team

        def on_submit(_):
            with self.output:
                clear_output()
                for team, btn in self.team_buttons.items():
                    took_action = btn.value and not btn.disabled
                    score = self.last_number if took_action else 0
                    self.team_scores[team].append(self.team_scores[team][-1] + score)
                    if took_action:
                        self.team_blocked_until[team] = self.current_round + self.last_cooldown
                update_plot()
            if self.current_round < self.max_rounds:
                start_new_round()
            else:
                df = (
                    pd.DataFrame(
                        {x: self.team_scores[x][-1] for x in self.team_scores}.items(),
                        columns=['Name', 'Score']
                    )
                    .sort_values('Score', ascending=False)
                    .reset_index(drop=True)
                    .rename(index=lambda x: x + 1)
                    .round(2)
                )
                self.number_display.value = ''
                self.submit_button.layout.display = 'none'
                for x in self.team_buttons.values():
                    x.layout.display = 'none'
                with self.output:
                    display(df)

        self.submit_button.on_click(on_submit)
        self.game_ui = widgets.HBox(
            [self.controls, self.plot_output],
            layout=widgets.Layout(
                align_items='center',
                justify_content='center',
                border='solid 1px gray',
                width='100%'
            )
        )

        top_spacer = widgets.Box(layout=widgets.Layout(flex='1 1 auto'))
        bottom_spacer = widgets.Box(layout=widgets.Layout(flex='1 1 auto'))
        self.vertical_box = widgets.VBox(
            [top_spacer, self.game_ui, bottom_spacer],
            layout=widgets.Layout(
                height='100vh',
                display='flex',
                flex_flow='column',
                align_items='center',
                justify_content='center',
            )
        )

        update_plot()
        start_new_round()

    def _ipython_display_(self):
        display(self.vertical_box)




import timeit
import numpy as np
import ipywidgets as widgets
import plotly.graph_objects as go
import pandas as pd
from IPython.display import clear_output, display
import time
import json

from .blackbox import generate_function


class CodingGame:
    def __init__(self, rounds, cooldown, plot_frequency=0.3, plot_width=1000, length_in_seconds=10, eliminate_slow_teams=False, functions=None, submissions=None):
        if functions is None and submissions is None:
            with open("submissions.json", "r") as f:
                submissions = json.load(f)
        self.candidates = functions if functions is not None else CodingGame.clean(submissions)
        self.plot_frequency = plot_frequency
        self.eliminate = eliminate_slow_teams
        self.length = length_in_seconds
        self.plot_width = plot_width
        self.last_cooldown = cooldown
        self.max_rounds = rounds
        self.init_play()
        self.initialize_ui()

    @staticmethod
    def clean(candidates):
        return {
            outer_key: generate_function(max(inner_dict.items(), key=lambda item: item[0])[1])
            for outer_key, inner_dict in candidates.items()
        }

    def initialize_ui(self):
        self.output = widgets.Output()
        self.plot_output = widgets.Output(layout=dict(width='100%'))
        self.game_ui = widgets.HBox(
            [self.output, self.plot_output],
            layout=widgets.Layout(
                align_items='center',
                justify_content='center',
                width='100%'
            )
        )
        top_spacer = widgets.Box(layout=widgets.Layout(flex='1 1 auto'))
        bottom_spacer = widgets.Box(layout=widgets.Layout(flex='1 1 auto'))
        self.vertical_box = widgets.VBox(
            [top_spacer, self.game_ui, bottom_spacer],
            layout=widgets.Layout(
                height='100vh',
                display='flex',
                flex_flow='column',
                align_items='center',
                justify_content='center',
            )
        )

        self.update_plot()

    def update_plot(self):
        fig = go.Figure()
        for team in self.candidates:
            n = self.current_round
            step = int(n / 100 + 1)
            fig.add_trace(go.Scatter(
                x=list(range(self.current_round + 1))[::step],
                y=self.team_scores[team][::step],
                mode='lines',
                name=team,
            ))
        fig.update_layout(showlegend=True, xaxis_title='Round', yaxis_title='Score', xaxis_range=[0, self.max_rounds], width=self.plot_width)
        with self.plot_output:
            clear_output(wait=True)
            display(fig)

    def start_new_round(self):
        self.current_round += 1
        self.last_number = self.gen.pareto(3)

    def loop(self):
        with self.output:
            while self.current_round < self.max_rounds:
                self.start_new_round()
                eliminate = []
                for team, btn in self.candidates.items():
                    args = (self.last_number, self.last_cooldown, self.current_round, self.max_rounds, self.team_scores[team][-1], [y[-1] for (x, y) in self.team_scores.items() if x != team])
                    tic_team = timeit.default_timer()
                    took_action = btn(*args) and self.current_round > self.team_blocked_until[team]
                    toc_team = timeit.default_timer()
                    if self.eliminate and toc_team - tic_team > self.length / self.max_rounds / len(self.candidates):
                        eliminate.append(team)
                    score = self.last_number if took_action else 0
                    self.team_scores[team].append(self.team_scores[team][-1] + score)
                    if took_action:
                        self.team_blocked_until[team] = self.current_round + self.last_cooldown
                for x in eliminate:
                    del self.candidates[x]
                toc = timeit.default_timer()
                target = self.start + self.current_round / self.max_rounds * self.length
                if toc < target:
                    time.sleep(target - toc)
                if toc > self.last_update + self.plot_frequency:
                    self.last_update = toc
                    self.update_plot()

            self.update_plot()
            df = (
                pd.DataFrame(
                    {x: self.team_scores[x][-1] for x in self.team_scores}.items(),
                    columns=['Name', 'Score']
                )
                .sort_values('Score', ascending=False)
                .reset_index(drop=True)
                .rename(index=lambda x: x + 1)
                .round(2)
            )
            display(df)

    def init_play(self):
        self.gen = np.random.default_rng(seed=None)
        self.last_update = -2 ** 31
        self.team_scores = {team: [0] for team in self.candidates}
        self.team_blocked_until = {team: -1 for team in self.candidates}
        self.last_number = None
        self.current_round = 0

    def play(self, countdown=0):

        with self.output:
            clear_output()
        with self.plot_output:
            clear_output()

        display(self.vertical_box)
        with self.output:
            for i in range(countdown, 0, -1):
                print(i)
                time.sleep(1)
                clear_output()
        self.start = timeit.default_timer()
        self.loop()

    def _ipython_display_(self):
        display(self.vertical_box)



from gr_tradinggame.manual import ManualGame

game = ManualGame(
    team_names=[
        'Watford',
        'Manchester',
        'Cambridge',
        'Edinburgh'
    ],
    max_rounds=3,
    cooldown=3,
    width=500
)



from gr_tradinggame.coding.client import Client

server = '28ca-45-80-244-212'

client = Client('Team A', server=server)

import random
def foo(reward: float, lockout: int, t: int, T: int, your_score: float, other_scores: list[float]):
    return random.random() > 0.5
client.test(foo)
client.test(foo, allow_state=True)

client.submit(foo, allow_state=True)

client = Client('Team B', server=server)

def foo(reward: float, lockout: int, t: int, T: int, your_score: float, other_scores: list[float]):
    threshold = -0.1
    if lockout <= 0:
        threshold = 0
    if t >= T:
        threshold = 0
    return reward > threshold
client.test(foo)
client.test(foo, strict=True)

def foo(reward: float, lockout: int, t: int, T: int, your_score: float, other_scores: list[float]):
    threshold = 0.1
    if lockout <= 0:
        threshold = 0
    if t >= T:
        threshold = 0
    return reward > threshold
client.test(foo, strict=True)

client.submit(foo)

client = Client('Team C', server=server)

client.submit('''
import time; 
def play(*args): time.sleep(0.1); return True
''', allow_state=True)

client.submit('''def play(*args): return True''', allow_state=True)

client = Client('Team D', server=server)

client.submit(foo)

def foo(reward: float, lockout: int, t: int, T: int, your_score: float, other_scores: list[float]):
    threshold = 0.2
    if lockout <= 0:
        threshold = 0
    if t >= T:
        threshold = 0
    return reward > threshold
client.submit(foo)



import base64
import datetime
import random
import traceback
import requests
from requests.auth import HTTPBasicAuth
import cloudpickle as pickle
import timeit

from .blackbox import Plackbox, Blackbox


class Client:
    def __init__(self, team_name, server=None, timeout=0.1):
        self.team_name = team_name
        if server:
            self.url = "https://" + server + ".ngrok-free.app"
        else:
            self.url = None
        self.timeout = timeout

    def test(self, source_or_function, verbose=False, strict=False, allow_state=False):
        return self._submit(source_or_function, strict=strict, submit=False, allow_state=allow_state, verbose=verbose)

    def submit(self, source_or_function, allow_state=False):
        return self._submit(source_or_function, strict=False, submit=True, allow_state=allow_state, verbose=False)
        
    def _submit(self, source_or_function, strict, submit, allow_state, verbose):
        if submit and not self.url:
            print('Automatic submission not enabled. Please send an email with your snippet to the organizers.')
            return
        test_cases = [
            ((1.5, 5, 1, 10, 1.3, [1.1, 2.3]), "generic inputs", None),
            ((-1.5, 5, 1, 10, 1.3, [1.1, 2.3]), "negative reward", (False, 'It never makes sense to take a negative reward')),
            ((0, 5, 1, 10, 1.3, [1.1, 2.3]), "zero rewards", (False, "It doesn't make sense to accept a lockout in return for no reward")),
            ((1.5, 5, 1, 10, 1.3, []), "playing by yourself", None),
            ((1.5, 5, 1, 10, -1, [-1.1, 2.3]), "negative score", None),
            ((0.0001, 5, 10, 10, 1.3, [1.1, 2.3]), "no time left after this round", (True, 'There is no time left after this round, you should accept any reward you can get')),
            ((0.0001, 0, 1, 10, 1.3, [1.1, 2.3]), "no lockout", (True, 'There is no lockout, you should accept any reward you can get')),
        ]
        if isinstance(source_or_function, str):
            output = source_or_function
            foo = Blackbox(output)
        else:
            output = pickle.dumps(source_or_function)
            foo = Plackbox(output)
        outputs = dict()
        for j in range(3):
            for i, (args, err, hint) in enumerate(test_cases):
                test_str = f'Test {i} ("{err}"): f{args}'
                if j == 0 and verbose:
                    print(test_str)
                    test_str = ''
                try:
                    a = foo(*args)
                    if j == 0:
                        outputs[i] = a
                    else:
                        if a != outputs[i] and not allow_state:
                            print(test_str)
                            raise Exception(f"""You previously returned `{outputs[i]}` and are now returning `{a}`. If this is intentional, set `allow_state=True`""")
                        elif not isinstance(a, bool):
                            print(test_str)
                            raise Exception(f"""You should return `False` or `True` but you returned `{a}` of type {type(a)}""")
                        elif strict and hint is not None and a != hint[0]:
                            print(test_str)
                            raise Exception(hint[1])
                except Exception:
                    print(test_str)
                    raise
                if j == 0 and verbose:
                    print(f'\tSuccess -- returned {a}')
        tic = timeit.default_timer()
        for j in range(1_000):
            T = int(1_000_000 * random.random())
            t = int(random.random() * T)
            cooldown = int(10 * random.random())
            reward = 100 * random.random()
            score = reward * T * random.random()
            scores = [reward * T * random.random() for _ in range(10)]
            args = (reward, cooldown, t, T,  score, scores)
            try:
                a = foo(*args)
            except Exception:
                print(f'Random input {args}')
                raise
            else:
                if not isinstance(a, bool):
                    print(f'Random input {args}')
                    raise Exception(f"""You should return `False` or `True` but returned `{a}` of type {type(a)}""")
            toc = timeit.default_timer()
            if toc - tic > max(1, (j + 1) * self.timeout):
                raise Exception(f"""Your code is too slow. Should take {self.timeout}s per round but took {(toc - tic) / (j + 1):.3f}s""")
        if submit:
            self.fancy_submit(source_or_function)
        else:
            print("All tests passed. You're ready to call `submit`\n")

    def fancy_submit(self, output):
        if isinstance(output, str):
            submission = 's' + base64.b64encode(output.encode('utf-8')).decode('utf-8')
        else:
            submission = 'p' + base64.b64encode(pickle.dumps(output)).decode('utf-8')
        j = {"team": self.team_name, 'time': str(datetime.datetime.now()), 'submission': submission}
        response = requests.post(
            self.url + '/receive',
            json=j,
            auth=HTTPBasicAuth("colabuser", "secretcolab"),
        )
        try:
            response.raise_for_status()
        except Exception:
            raise Exception("Couldn't connect to server. Please notify an organizer.")
        else:
            print('Submission successful: ', j['team'], j['time'])
